/* PE.c - functions to work with Win32 Portable Executable files */

#include <windows.h>
#include <stddef.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>


extern int gDiag;


/* Check the DOS and PE headers and validate them. */
static long validateNTHeaders(FILE *target, IMAGE_DOS_HEADER *dosHdr,
							  IMAGE_NT_HEADERS32 *ntHdr)
{
	size_t  count;
	IMAGE_DOS_HEADER   tempDosHdr;
	IMAGE_NT_HEADERS32 tempNtHdr;

	if (dosHdr == NULL)
		dosHdr = &tempDosHdr;

	if (ntHdr == NULL)
		ntHdr = &tempNtHdr;

	if (fseek(target, 0, SEEK_SET))
	{
		fprintf(stderr, "ERROR: Could not seek. Is file open?\n");
		return(-1);
	}

	if (((count = fread(dosHdr, 1, sizeof(IMAGE_DOS_HEADER), target)) == EOF) ||
		 (count < sizeof(IMAGE_DOS_HEADER)))
	{
		fprintf(stderr, "ERROR: Could not read %d bytes\n", sizeof(IMAGE_DOS_HEADER));
		return(-1);
	}

	if (dosHdr->e_magic != IMAGE_DOS_SIGNATURE)
	{
		fprintf(stderr, "ERROR: The target does not appear to be an executable file\n");
		return(-1);
	}

	if (fseek(target, dosHdr->e_lfanew, SEEK_SET))
	{
		fprintf(stderr, "ERROR: Could not seek to offset %d\n", dosHdr->e_lfanew);
		return(-1);
	}

	if ((count = fread(ntHdr, 1, sizeof(IMAGE_NT_HEADERS32), target)) == EOF ||
		 count < sizeof(IMAGE_NT_HEADERS32))
	{
		fprintf(stderr, "ERROR: Could not read NT headers\n");
		return(-1);
	}

	if (ntHdr->FileHeader.Machine != IMAGE_FILE_MACHINE_I386)
	{
		fprintf(stderr, "ERROR: File is not valid for Intel i386 processors\n");
		return(-1);
	}

	if (ntHdr->Signature != IMAGE_NT_SIGNATURE)
	{
		fprintf(stderr, "ERROR: File is not a valid NT executable file\n");
		return(-1);
	}

	/* Return the file offset to the PE header */
	return((long)dosHdr->e_lfanew);
}



/* Update RVAs in the resource data entries. */
static void updateResEntryRVA(DWORD newResBase, DWORD oldResBase, char *pNewRes,
					   IMAGE_RESOURCE_DIRECTORY *pResDir)
{
	int i;
    IMAGE_RESOURCE_DIRECTORY_ENTRY *pResDirEntry = NULL;
	IMAGE_RESOURCE_DATA_ENTRY      *pResDataEntry = NULL;

	pResDirEntry = (IMAGE_RESOURCE_DIRECTORY_ENTRY *)&pResDir[1];

	for (i = 0; i < pResDir->NumberOfNamedEntries + pResDir->NumberOfIdEntries; i++)
	{
		if (pResDirEntry[i].DataIsDirectory) /* Recurse if directory */
		{
			updateResEntryRVA(newResBase, oldResBase, pNewRes,
							(IMAGE_RESOURCE_DIRECTORY *)((char *)pNewRes +
							pResDirEntry[i].OffsetToDirectory));
		}
		else
		{
			pResDataEntry = (IMAGE_RESOURCE_DATA_ENTRY *)(pNewRes +
							pResDirEntry[i].OffsetToData);
			pResDataEntry->OffsetToData = pResDataEntry->OffsetToData -
										  oldResBase + newResBase;
		}
	}
}



/* Check to see if a .rsrc section exists in the target executable file.   */
int resourceSectionExists(wchar_t *fileName)
{
	int i;
	size_t  count;
	FILE   *targetFile;
	IMAGE_DOS_HEADER     dosHdr; /* From winnt.h */
	IMAGE_NT_HEADERS32   ntHdrs; /* From winnt.h */
	IMAGE_SECTION_HEADER secHdr; /* From winnt.h */

	if ((targetFile = _wfopen(fileName, L"rb")) == NULL)
	{
		fprintf(stderr, "ERROR: Cannot open %S\n", fileName);
		return(0);
	}

	if (validateNTHeaders(targetFile, &dosHdr, &ntHdrs) < 0)
	{
		fprintf(stderr, "ERROR: Could not validate PE headers in %S\n", fileName);
	    fclose(targetFile);
		return(0);
	}

	for (i = 0; i < ntHdrs.FileHeader.NumberOfSections; i++)
	{
	    if ((count = fread(&secHdr, 1, sizeof(secHdr), targetFile)) == EOF ||
		     count < sizeof(secHdr))
		{
		    fprintf(stderr, "ERROR: Could not read SectionHeader[%d]from %S\n",
			                 i, fileName);
	        fclose(targetFile);
		    return(0);
		}

		if (strncmp(secHdr.Name, ".rsrc", 5) == 0)
		{
	        fclose(targetFile);
			return(1);
		}

	}


	fclose(targetFile);
	return(0);
}



/* Add a .rsrc section to an executable file that does not have one. */
int addResourceSection(wchar_t *fileName)
{
	int  i;
	int  j;
	int  dbgDirCount = 0;           /* How many debug directories?     */
	long ntHeader;                  /* File offset of NT headers       */
	long resourceSectionTableEntry; /* File offset for section header  */
	DWORD resSectRVA;               /* RVA of resource section         */
	DWORD resSectFileOffset;         /* File offset of resource section */
	DWORD dbgDirOffset = 0;          /* File offset of debug directory  */
	long endOfFile;                 /* Offset of last byte of file     */
	long appendedBytes = 0;         /* Size appended after last section*/
	long newResSize = 0;            /* Size of target's .rsrc section  */
	long myResSize = 0;             /* Size of this programs resources */
	char *appendBuff = NULL;        /* malloc'ed buffer, appended data */
	char *newRes = NULL;            /* malloc'ed copy of myResources   */
	char *myResources;              /* Pointer, this process resources */
	size_t  count;
	FILE   *targetFile = NULL;
	IMAGE_DOS_HEADER     dosHdr;    /* Target file, from winnt.h  */
	IMAGE_NT_HEADERS32   ntHdrs;    /* Target file, from winnt.h  */
	IMAGE_SECTION_HEADER secHdr;    /* Target file, from winnt.h  */
	IMAGE_DEBUG_DIRECTORY *dbgDir = NULL; /* Target file, from winnt.h */
	IMAGE_SECTION_HEADER *sectionHeaders = NULL;
	IMAGE_DOS_HEADER     *myDosHdr; /* This process, from winnt.h */
	IMAGE_NT_HEADERS32   *myNtHdrs; /* This process, from winnt.h */
	IMAGE_SECTION_HEADER *myResHdr; /* This process, from winnt.h */

	if ((targetFile = _wfopen(fileName, L"r+b")) == NULL)
	{
		fprintf(stderr, "ERROR: Cannot open %S\n", fileName);
		return(0);
	}

	if ((ntHeader = validateNTHeaders(targetFile, &dosHdr, &ntHdrs)) < 0)
	{
		fprintf(stderr, "ERROR: Could not validate PE headers in %S\n", fileName);
		goto myErrorExit;
	}


	sectionHeaders = (IMAGE_SECTION_HEADER *)malloc(ntHdrs.FileHeader.NumberOfSections * 
		                    sizeof(IMAGE_SECTION_HEADER));

	secHdr.Name[0] = '\0';

	for (i = 0; i < ntHdrs.FileHeader.NumberOfSections; i++)
	{
	    if ((count = fread(&secHdr, 1, sizeof(secHdr), targetFile)) == EOF ||
		     count < sizeof(secHdr))
		{
		    fprintf(stderr, "ERROR: Could not read SectionHeader[%d]from %S\n",
			                 i, fileName);
			goto myErrorExit;
		}

		memcpy(&sectionHeaders[i], &secHdr, sizeof(secHdr));

		if (strncmp(secHdr.Name, ".rsrc", 5) == 0)
		{
			free(sectionHeaders);
	        fclose(targetFile);
			return(1);
		}

	}

	/* If we got here, it means the file does not yet have a .rsrc section  */
	if (secHdr.Name[0] == '\0')
	{
		fprintf(stderr, "ERROR: Invalid section header.\n");
		goto myErrorExit;
	}

	if (gDiag)
		fprintf(stderr, "\t\tAdding resource section.\n");

	resSectRVA = ntHdrs.OptionalHeader.SizeOfImage;
	resSectFileOffset = secHdr.PointerToRawData + secHdr.SizeOfRawData;
	resourceSectionTableEntry = ftell(targetFile);

	if (resourceSectionTableEntry + sizeof(secHdr) >= ntHdrs.OptionalHeader.SizeOfHeaders)
	{
		fprintf(stderr, "Cannot add .rsrc section, not enough space in table.\n");
		goto myErrorExit; /* Clean up and return an error indication */
	}

	if ((myDosHdr = (IMAGE_DOS_HEADER *)GetModuleHandle(NULL)) == NULL)
	{
		fprintf(stderr, "Cannot get load address (myDosHdr) for this process.\n");
		goto myErrorExit;
	}

	myNtHdrs = (IMAGE_NT_HEADERS32 *)((char *)myDosHdr + myDosHdr->e_lfanew);
	myResources = (char *)myDosHdr + myNtHdrs->OptionalHeader.
		          DataDirectory[IMAGE_DIRECTORY_ENTRY_RESOURCE].VirtualAddress;
	myResSize = myNtHdrs->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_RESOURCE].Size;
	myResHdr = (IMAGE_SECTION_HEADER *)((char *)myNtHdrs + sizeof(IMAGE_NT_HEADERS32));

	for (i = 0; i < myNtHdrs->FileHeader.NumberOfSections; i++)
	{
		myResHdr = &myResHdr[i];

		if (!strncmp(myResHdr->Name, ".rsrc", 5))
			break;
	}

	for (i = 0; i < IMAGE_NUMBEROF_DIRECTORY_ENTRIES; i++)
	{
		if (ntHdrs.OptionalHeader.DataDirectory[i].VirtualAddress != 0 ||
			ntHdrs.OptionalHeader.DataDirectory[i].Size != 0)
		{
			switch(i)
			{
			case IMAGE_DIRECTORY_ENTRY_SECURITY:
				if (ntHdrs.OptionalHeader.DataDirectory[i].VirtualAddress >= resSectFileOffset)
				{
					fprintf(stderr, "ERROR: Can't handle DIRECTORY_ENTRY_SECURITY\n");
					goto myErrorExit;
				}

				break;

			case IMAGE_DIRECTORY_ENTRY_DEBUG:
				if (gDiag)
					fprintf(stderr, "\t\tDebug data found.\n");

				dbgDirCount = ntHdrs.OptionalHeader.DataDirectory[i].Size /
					          sizeof(IMAGE_DEBUG_DIRECTORY);

				for (j = 0; j < ntHdrs.FileHeader.NumberOfSections; j++)
				{
					if (ntHdrs.OptionalHeader.DataDirectory[i].VirtualAddress >=
						sectionHeaders[j].VirtualAddress &&
						ntHdrs.OptionalHeader.DataDirectory[i].VirtualAddress <
						sectionHeaders[j].VirtualAddress +
						sectionHeaders[j].Misc.VirtualSize)
					{
						dbgDirOffset = sectionHeaders[j].PointerToRawData +
							     (ntHdrs.OptionalHeader.DataDirectory[i].VirtualAddress - 
								 sectionHeaders[j].VirtualAddress);
						break;
					}
				}

				break;

			case IMAGE_DIRECTORY_ENTRY_EXPORT:
			case IMAGE_DIRECTORY_ENTRY_IMPORT:
			case IMAGE_DIRECTORY_ENTRY_RESOURCE:
			case IMAGE_DIRECTORY_ENTRY_EXCEPTION:
			case IMAGE_DIRECTORY_ENTRY_BASERELOC:
			case IMAGE_DIRECTORY_ENTRY_ARCHITECTURE:
			case IMAGE_DIRECTORY_ENTRY_GLOBALPTR:
			case IMAGE_DIRECTORY_ENTRY_TLS:
			case IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG:
			case IMAGE_DIRECTORY_ENTRY_BOUND_IMPORT:
			case IMAGE_DIRECTORY_ENTRY_IAT:
			case IMAGE_DIRECTORY_ENTRY_DELAY_IMPORT:
			case IMAGE_DIRECTORY_ENTRY_COM_DESCRIPTOR:
			default:
				if (ntHdrs.OptionalHeader.DataDirectory[i].VirtualAddress >= resSectRVA)
				{
					fprintf(stderr, "ERROR: Can't handle directory entry %d\n", i);
					goto myErrorExit;
				}

				break;
			}
		}
	}

	newResSize = ntHdrs.OptionalHeader.FileAlignment;

	if (newResSize < myResSize)
		newResSize *= (myResSize / newResSize + 1);

	/* Check if there is anything (debug info?) appended to the    */
	/* file that might be overwritten by the new resource section. */
	fseek(targetFile,0, SEEK_END);
	endOfFile = ftell(targetFile);
	fseek(targetFile, (long)resSectFileOffset, SEEK_SET);

	if (endOfFile >= resSectFileOffset)
	{
		appendedBytes = endOfFile - resSectFileOffset;

		if ((appendBuff = (char *)malloc(appendedBytes)) == NULL)
		{
			fprintf(stderr, "ERROR: Could not malloc() %ld byte buffer\n", appendedBytes);
			goto myErrorExit;
		}

		if ((count = fread(appendBuff, 1, appendedBytes, targetFile)) < appendedBytes)
		{
			fprintf(stderr, "ERROR: Could not read %ld bytes into buffer\n", appendedBytes);
			goto myErrorExit;
		}

		if (gDiag)
			fprintf(stderr, "\t\t\t%ld bytes of data (debug info) saved.\n", appendedBytes);
	}

	/* If there are debug directories, read and correct them */
	if (dbgDirOffset)
	{
		fseek(targetFile, (long)dbgDirOffset, SEEK_SET);

		if ((dbgDir = (IMAGE_DEBUG_DIRECTORY *)malloc(sizeof(IMAGE_DEBUG_DIRECTORY) *
			                                          dbgDirCount)) == NULL)
		{
			fprintf(stderr, "ERROR: Could not malloc() dbgDir buffer\n");
			goto myErrorExit;
		}

		if ((count = fread(dbgDir, 1, sizeof(IMAGE_DEBUG_DIRECTORY) * dbgDirCount,
			         targetFile)) < sizeof(IMAGE_DEBUG_DIRECTORY) * dbgDirCount)
		{
			fprintf(stderr, "ERROR: Could not read debugDirectory\n");
			goto myErrorExit;
		}

		/* Seek back so we can update the structures */
		fseek(targetFile, (long)dbgDirOffset, SEEK_SET);

		for (i = 0; i < dbgDirCount; i++)
		{
			/* If the debug data is appended after the last section, change */
			/* the offset in the debug directory so it can be located.      */
			if (dbgDir[i].PointerToRawData >= resSectFileOffset)
				dbgDir[i].PointerToRawData += newResSize;

			if (dbgDir[i].AddressOfRawData >= resSectRVA)
			{
				fprintf(stderr, "ERROR: Can't handle dbgDir.AddressOfRawData > 0\n");
				goto myErrorExit;
			}
		}

		if (fwrite(dbgDir, 1, sizeof(IMAGE_DEBUG_DIRECTORY) * dbgDirCount, targetFile) <
			 (sizeof(IMAGE_DEBUG_DIRECTORY) * dbgDirCount))
		{
			fprintf(stderr, "ERROR: Could not write updated debug directory\n");
			goto myErrorExit;
		}

		free(dbgDir);
		dbgDir = NULL;

		if (gDiag)
			fprintf(stderr, "\t\t\t%d debug directories found & updated.\n", dbgDirCount);
	}

	fseek(targetFile, (long)resSectFileOffset, SEEK_SET);

	if ((newRes = (char *)calloc(newResSize, 1)) == NULL)
	{
		fprintf(stderr, "ERROR: Could not calloc() newRes buffer\n");
		goto myErrorExit;
	}

	memcpy(newRes, myResources, myResSize);
	updateResEntryRVA(resSectRVA, myResHdr->VirtualAddress, newRes,
		              (IMAGE_RESOURCE_DIRECTORY *)newRes);

	if (gDiag)
		fprintf(stderr, "\t\tUpdated offsets in cloned resource section.\n");

	if ((count = fwrite(newRes, 1, newResSize, targetFile)) < newResSize)
	{
		fprintf(stderr, "ERROR: Could not write newRes section\n");
		goto myErrorExit;
	}

	free(newRes);
	newRes = NULL;

	if (gDiag)
		fprintf(stderr, "\t\tSuccessfully wrote cloned resource section.\n");

	if (appendBuff)
	{
		if ((count = fwrite(appendBuff, 1, appendedBytes, targetFile)) < appendedBytes)
		{
			fprintf(stderr, "ERROR: Could not write appended stuff\n");
			goto myErrorExit;
		}

		free(appendBuff);
		appendBuff = NULL;

		if (gDiag)
			fprintf(stderr, "\t\tSuccessfully wrote appended data (debug info).\n");
	}

	fseek(targetFile, resourceSectionTableEntry, SEEK_SET);

	/* Initialize our section header for the new resource section  */
	strncpy(secHdr.Name, ".rsrc", 8);
	secHdr.Misc.VirtualSize = myResSize;
	secHdr.VirtualAddress = resSectRVA;
	secHdr.SizeOfRawData = newResSize;
	secHdr.PointerToRawData = resSectFileOffset;
	secHdr.PointerToRelocations = 0;
	secHdr.PointerToLinenumbers = 0;
	secHdr.NumberOfRelocations = 0;
	secHdr.NumberOfLinenumbers = 0;
	secHdr.Characteristics = myResHdr->Characteristics;

	fwrite(&secHdr, sizeof(secHdr), 1, targetFile);

	if (gDiag)
		fprintf(stderr, "\t\tUpdated resource section header.\n");

	ntHdrs.FileHeader.NumberOfSections++;
	ntHdrs.OptionalHeader.SizeOfInitializedData += newResSize;
	ntHdrs.OptionalHeader.SizeOfImage += newResSize;

	/* Checksum is only required for drivers. */
	if (ntHdrs.OptionalHeader.CheckSum)
	{
		ntHdrs.OptionalHeader.CheckSum = 0;
		fprintf(stderr, "WARNING: Checksum set to 0. If %S is a driver, the\n\t"
				"checksum needs to be recalculated with EDITBIN /RELEASE\n", fileName);
	}

	ntHdrs.OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_RESOURCE].VirtualAddress = resSectRVA;
	ntHdrs.OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_RESOURCE].Size = myResSize;
	fseek(targetFile, ntHeader, SEEK_SET);
	fwrite(&ntHdrs, sizeof(ntHdrs), 1, targetFile);

	if (gDiag)
		fprintf(stderr, "\t\tUpdated PE header offsets, sizes, etc.\n");

	free(sectionHeaders);
	fclose(targetFile);

	return(1); /* Indicate success */


/* Emergency exit when things go south. */
myErrorExit:
	if (newRes)
		free(newRes);

	if (dbgDir)
		free(dbgDir);

	if (appendBuff)
		free(appendBuff);

	if (sectionHeaders)
		free(sectionHeaders);

	if (targetFile)
		fclose(targetFile);

	return(0);
}

